#! /usr/bin/env perl -w

# Define Setmanager class
package Setmanager;

sub new {
    my ($class,$mod) = @_;
    my ($self) = {};
	bless $self, $class;	
	$self->{'_mod'} = $mod; #assign empty list	    
    return $self;
}

sub setMod($$) {
    my ( $self, $mod ) = @_;
    $self->{_mod} = $mod;
    return $self->{_mod};
}

sub invert {
    my( $self, $pcsRef ) = @_;
    my @pcs = sort(@{$pcsRef});
    my @inv;
    foreach (@pcs) { push(@inv,$_*-1); }
    @inv = sort(@inv);
    return \@inv;
}

sub modular {
    my( $self, $n ) = @_;
    while ($n < 0) { $n += $self->{'_mod'}; }
	while ($n >= $self->{'_mod'}) { $n -= $self->{'_mod'}; }
	return $n;
}

sub getTnclass {
	my( $self, $pcsRef ) = @_;
	my @pcs = @{$pcsRef};
	foreach my $p (@pcs) { $p = $self->modular($p); }
	@pcs = sort {$a<=>$b} @pcs;
	my @tnclass = @pcs;
	for (my $k = 0; $k < @pcs; $k++)
    {
    	$pcs[0] += $self->{'_mod'};
     	@pcs = sort {$a<=>$b} @pcs;
	   	for (my $i = $#pcs; $i > 0; $i--)
    	{
			my $p = $pcs[$i]-$pcs[0];
			my $t = $tnclass[$i]-$tnclass[0];
    		if ($p == $t) { next; }
    		elsif ($p < $t) { @tnclass = @pcs; }
    		last;
    	}
    }
    
	my $diff = $tnclass[0];
    for (my $k = 0; $k < @tnclass; $k++)
    {
		$tnclass[$k] -= $diff;
	}
    return \@tnclass;	
}

sub getSetclass {
	my ( $self, $pcsRef ) = @_;
	my @pcs = sort {$a<=>$b} @{$pcsRef};
	my @tnclass = @{$self->getTnclass($pcsRef)};
	my $invref = $self->invert($pcsRef);
	my @inv = @{$self->getTnclass($invref)};
	my @setclass = @tnclass;
	for (my $k = @tnclass-1; $k >= 0; $k--)
	{
		my $i = $inv[$k]-$inv[0];
		my $t = $tnclass[$k]-$tnclass[0];
		if ($i > $t) { last; }
		elsif ($i < $t) { @setclass = @inv; last; }
	}
	return \@setclass;
}

sub getIntervals {
	my ( $self, $pcsRef ) = @_;
	my @pcs = sort {$a<=>$b} @{$pcsRef};
	my @ints;
	for (my $i = 0; $i <= $#pcs; $i++)
	{
		for (my $j = $i+1; $j <= $#pcs; $j++)
		{
			my $int = $pcs[$j]-$pcs[$i];
			$int = $self->modular($int);
			if ($int > $self->{'_mod'}/2) { 
				$int = $self->{'_mod'} - $int; 
			}
			$int = sprintf("%.5f",$int);
			$int =~ s/0+$//;
			$int =~ s/\.$//;
			push(@ints,$int);
		}
	}
	@ints = sort(@ints);
	return \@ints;
}

1;

package main;
# End Setmanager definition


# Some Extra Methods
# ---------------------------------------------
sub isNumber {
  my $ss = shift;
  return 1 if $ss =~ /^[+\-]?\d*\.?\d+$/;
  return 1 if $ss =~ /^[+\-]?\d*\.?\d+e[+\-]?\d+$/i;
  return 0;
}

sub formatFloat {
	my ($n) = @_;
	$n = sprintf("%.5f",$n);
	$n =~ s/0+$//;
	$n =~ s/\.$//;
	return $n;
}

sub usage {
	print "**********************************************\n";
	print "Usage: setclass\n";
	print "Usage: setclass [mod] [pcs...]\n";
	print "Example: setclass 12 0 1 4 6\n";
	print "**********************************************\n";
	exit();	
}

# Main part
# ---------------------------------------------
my $mod = 'a';
my @pcs;
if (@ARGV < 1) {
	while (!isNumber($mod)) {
		print "Enter a mod: ";
		$mod = <>;
	}
	print "Enter pcs: ";
	my $pcstr = <>;
	$pcstr =~ s/\s+/ /;
	my @temp = split(' ', $pcstr);
	foreach my $p (@temp) 
	{
		if (isNumber($p)) { push(@pcs,$p); }
	}
}
else {
	$mod = shift @ARGV;
	if (!isNumber($mod)) { usage(); }
	#my @pcs = [6.6,7,3,1];
	foreach $a (@ARGV) { if (isNumber($a)) { push(@pcs,"$a"); }}
}

if (!@pcs) { usage(); }

# Now that $mod and @pcs are defined, do the stuff...
my $pcsref = \@pcs;
my $object = new Setmanager($mod);
my @tnclass = @{$object->getTnclass($pcsref)};
my @setclass = @{$object->getSetclass($pcsref)};
my @intervals = @{$object->getIntervals(\@setclass)};

print "Tnclass: ";
foreach my $t (@tnclass) { print formatFloat($t) . " "; }
print "\n";
print "Setclass: ";
foreach my $s (@setclass) { print formatFloat($s) . " "; }
my $c = 1;
print "\nIntervals: ";
for (my $i = 0; $i <= $#intervals; $i++)
{
	if ($i != $#intervals && $intervals[$i] == $intervals[$i+1]) { 
		$c++; next; 
	}
	print $intervals[$i] . "-" . $c . " ";
	$c=1;
}
print "\n";
